import itertools
import numpy as np
import discretisedfield as df
import discretisedfield.util as dfu
from scipy import ndimage


def topological_charge_density(field, /, method='continuous'):
    r"""Topological charge density.

    This method computes the topological charge density for a vector field
    (``dim=3``). Two different methods are available and can be selected using
    ``method``:

    1. Continuous method:

        .. math::

            q = \frac{1}{4\pi} \mathbf{n} \cdot \left(\frac{\partial
            \mathbf{n}}{\partial x} \times \frac{\partial
            \mathbf{n}}{\partial x} \right),

        where :math:`\mathbf{n}` is the orientation field.

    2. Berg-Luescher method. Details can be found in:

        1. B. Berg and M. Luescher. Definition and statistical distributions of
        a topological number in the lattice O(3) sigma-model. Nuclear Physics B
        190 (2), 412-424 (1981).

        2. J.-V. Kim and J. Mulkers. On quantifying the topological charge in
        micromagnetics using a lattice-based approach. IOP SciNotes 1, 025211
        (2020).

    Topological charge is defined on two-dimensional samples only. Therefore,
    the field must be "sliced" using the ``discretisedfield.Field.plane``
    method. If the field is not three-dimensional or the field is not sliced,
    ``ValueError`` is raised.

    Parameters
    ----------
    field : discretisedfield.Field

        Vector field.

    method : str, optional

        Method how the topological charge is computed. It can be ``continuous``
        or ``berg-luescher``. Defaults to ``continuous``.

    Returns
    -------
    discretisedfield.Field

        Topological charge density scalar field.

    Raises
    ------
    ValueError

        If the field is not three-dimensional or the field is not sliced.

    Example
    -------
    1. Compute topological charge density of a spatially constant vector field.

    >>> import discretisedfield as df
    >>> import discretisedfield.tools as dft
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (10, 10, 10)
    >>> cell = (2, 2, 2)
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
    >>> f = df.Field(mesh, dim=3, value=(1, 1, -1))
    ...
    >>> dft.topological_charge_density(f.plane('z'))
    Field(...)
    >>> dft.topological_charge_density(f.plane('z'), method='berg-luescher')
    Field(...)

    2. An attempt to compute the topological charge density of a scalar field.

    >>> f = df.Field(mesh, dim=1, value=12)
    >>> dft.topological_charge_density(f.plane('z'))
    Traceback (most recent call last):
    ...
    ValueError: ...

    3. Attempt to compute the topological charge density of a vector field,
    which is not sliced.

    >>> f = df.Field(mesh, dim=3, value=(1, 2, 3))
    >>> dft.topological_charge_density(f)
    Traceback (most recent call last):
    ...
    ValueError: ...

    .. seealso:: :py:func:`~discretisedfield.tools.topological_charge`

    """
    if field.dim != 3:
        msg = (f'Cannot compute topological charge density '
               f'for {field.dim=} field.')
        raise ValueError(msg)

    if not field.mesh.attributes['isplane']:
        msg = ('The field must be sliced before the topological '
               'charge density can be computed.')
        raise ValueError(msg)

    if method not in ['continuous', 'berg-luescher']:
        msg = 'Method can be either continuous or berg-luescher'
        raise ValueError(msg)

    axis1 = field.mesh.attributes['axis1']
    axis2 = field.mesh.attributes['axis2']
    of = field.orientation  # unit field - orientation field

    if method == 'continuous':
        return 1/(4*np.pi) * of @ (of.derivative(dfu.raxesdict[axis1]) &
                                   of.derivative(dfu.raxesdict[axis2]))

    elif method == 'berg-luescher':
        q = field.__class__(field.mesh, dim=1, value=0)

        # Area of a single triangle
        area = 0.5 * field.mesh.cell[axis1] * field.mesh.cell[axis2]

        for i, j in itertools.product(range(of.mesh.n[axis1]),
                                      range(of.mesh.n[axis2])):
            index = dfu.assemble_index(0, 3, {axis1: i, axis2: j})
            v0 = of.array[index]

            # Extract 4 neighbouring vectors (if they exist)
            v1 = v2 = v3 = v4 = None
            if i + 1 < of.mesh.n[axis1]:
                v1 = of.array[dfu.assemble_index(0, 3, {axis1: i+1,
                                                        axis2: j})]
            if j + 1 < of.mesh.n[axis2]:
                v2 = of.array[dfu.assemble_index(0, 3, {axis1: i,
                                                        axis2: j+1})]
            if i - 1 >= 0:
                v3 = of.array[dfu.assemble_index(0, 3, {axis1: i-1,
                                                        axis2: j})]
            if j - 1 >= 0:
                v4 = of.array[dfu.assemble_index(0, 3, {axis1: i,
                                                        axis2: j-1})]

            charge = 0
            triangle_count = 0

            if v1 is not None and v2 is not None:
                triangle_count += 1
                charge += dfu.bergluescher_angle(v0, v1, v2)

            if v2 is not None and v3 is not None:
                triangle_count += 1
                charge += dfu.bergluescher_angle(v0, v2, v3)

            if v3 is not None and v4 is not None:
                triangle_count += 1
                charge += dfu.bergluescher_angle(v0, v3, v4)

            if v4 is not None and v1 is not None:
                triangle_count += 1
                charge += dfu.bergluescher_angle(v0, v4, v1)

            if triangle_count > 0:
                q.array[index] = charge / (area * triangle_count)
            else:
                # If the cell has no neighbouring cells
                q.array[index] = 0

        return q


def topological_charge(field, /, method='continuous', absolute=False):
    """Topological charge.

    This function computes topological charge for a vector field (``dim=3``).
    There are two possible methods, which can be chosen using ``method``
    parameter. For details on method, please refer to
    :py:func:`~discretisedfield.tools.topological_charge_density`. Absolute
    topological charge given as integral over the absolute values of the
    topological charge density can be computed by passing ``absolute=True``.

    Topological charge is defined on two-dimensional samples. Therefore,
    the field must be "sliced" using ``discretisedfield.Field.plane``
    method. If the field is not three-dimensional or the field is not
    sliced and ``ValueError`` is raised.

    Parameters
    ----------
    field : discretisedfield.Field

        Vector field.

    method : str, optional

        Method how the topological charge is computed. It can be ``continuous``
        or ``berg-luescher``. Defaults to ``continuous``.

    absolute : bool, optional

        If ``True`` the absolute topological charge is computed.
        Defaults to ``False``.

    Returns
    -------
    float

        Topological charge.

    Raises
    ------
    ValueError

        If the field is not three-dimensional or the field is not sliced.

    Example
    -------
    1. Compute the topological charge of a spatially constant vector field
    (zero value is expected).

    >>> import discretisedfield as df
    >>> import discretisedfield.tools as dft
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (10, 10, 10)
    >>> cell = (2, 2, 2)
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
    ...
    >>> f = df.Field(mesh, dim=3, value=(1, 1, -1))
    >>> dft.topological_charge(f.plane('z'), method='continuous')
    0.0
    >>> dft.topological_charge(f.plane('z'), method='continuous',
    ...                                      absolute=True)
    0.0
    >>> dft.topological_charge(f.plane('z'), method='berg-luescher')
    0.0
    >>> dft.topological_charge(f.plane('z'), method='berg-luescher',
    ...                                      absolute=True)
    0.0

    2. Attempt to compute the topological charge of a scalar field.

    >>> f = df.Field(mesh, dim=1, value=12)
    >>> dft.topological_charge(f.plane('z'))
    Traceback (most recent call last):
    ...
    ValueError: ...

    3. Attempt to compute the topological charge of a vector field, which
    is not sliced.

    >>> f = df.Field(mesh, dim=3, value=(1, 2, 3))
    >>> dft.topological_charge(f)
    Traceback (most recent call last):
    ...
    ValueError: ...

    .. seealso::

        :py:func:`~discretisedfield.tools.topological_charge_density`

    """
    if field.dim != 3:
        msg = f'Cannot compute topological charge for {field.dim=} field.'
        raise ValueError(msg)

    if not field.mesh.attributes['isplane']:
        msg = ('The field must be sliced before the '
               'topological charge can be computed.')
        raise ValueError(msg)

    if method not in ['continuous', 'berg-luescher']:
        msg = 'Method can be either continuous or berg-luescher'
        raise ValueError(msg)

    if method == 'continuous':
        q = topological_charge_density(field, method=method)
        if absolute:
            return df.integral(abs(q) * abs(df.dS))
        else:
            return df.integral(q * abs(df.dS))

    elif method == 'berg-luescher':
        axis1 = field.mesh.attributes['axis1']
        axis2 = field.mesh.attributes['axis2']
        of = field.orientation

        topological_charge = 0
        for i, j in itertools.product(range(of.mesh.n[axis1] - 1),
                                      range(of.mesh.n[axis2] - 1)):
            v1 = of.array[dfu.assemble_index(0, 3, {axis1: i,
                                                    axis2: j})]
            v2 = of.array[dfu.assemble_index(0, 3, {axis1: i + 1,
                                                    axis2: j})]
            v3 = of.array[dfu.assemble_index(0, 3, {axis1: i + 1,
                                                    axis2: j + 1})]
            v4 = of.array[dfu.assemble_index(0, 3, {axis1: i,
                                                    axis2: j + 1})]

            triangle1 = dfu.bergluescher_angle(v1, v2, v4)
            triangle2 = dfu.bergluescher_angle(v2, v3, v4)

            if absolute:
                triangle1 = abs(triangle1)
                triangle2 = abs(triangle2)

            topological_charge += triangle1 + triangle2

        return topological_charge


def emergent_magnetic_field(field):
    r"""Emergent magnetic field.

    Emergent magnetic field for a (magnetic) unit vector field
    :math:`\boldsymbol{m}`is defined as:

    .. math::

        F_{kl} = \boldsymbol{m} \cdot (\partial_k \boldsymbol{m}
        \times \partial_l \boldsymbol{m})

    Details are given in Volovik, G. E., Rysti, J., Mäkinen, J. T. & Eltsov,
    V. B. Spin, Orbital, Weyl and Other Glasses in Topological Superfluids. J
    Low Temp Phys 196, 82–101 (2019).

    Parameters
    ----------
    field : discretisedfield.Field

        Vector field.

    Returns
    -------
    discretisedfield.Field

        Emergent magnetic field.

    Raises
    ------
    ValueError

        If the field is not three-dimensional.

    Example
    -------
    1. Compute topological charge density of a spatially constant vector field.

    >>> import discretisedfield as df
    >>> import discretisedfield.tools as dft
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (10, 10, 10)
    >>> cell = (2, 2, 2)
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
    >>> f = df.Field(mesh, dim=3, value=(1, 1, -1))
    ...
    >>> dft.emergent_magnetic_field(f)
    Field(...)

    """
    if field.dim != 3:
        msg = (f'Cannot compute emergent magnetic field '
               f'for {field.dim=} field.')
        raise ValueError(msg)

    Fx = field @ (field.derivative('y') & field.derivative('z'))
    Fy = field @ (field.derivative('z') & field.derivative('x'))
    Fz = field @ (field.derivative('x') & field.derivative('y'))

    return Fx << Fy << Fz


def neigbouring_cell_angle(field, /, direction, units='rad'):
    """Calculate angles between neighbouring cells.

    This method calculates the angle between magnetic moments in all
    neighbouring cells. The calculation is only possible for vector fields
    (``dim=3``). Angles are computed in degrees if ``units='deg'`` and in
    radians if ``units='rad'``.

    The resulting field has one discretisation cell less in the specified
    direction.

    Parameters
    ----------
    field : discretisedfield.Field

        Vector field.

    direction : str

        The direction in which the angles are calculated. Can be ``'x'``,
        ``'y'`` or ``'z'``.

    units : str, optional

        Angles are computed in degrees if ``units='deg'`` and in radians if
        ``units='rad'``. Defaults to ``'rad'``.

    Returns
    -------
    discretisedfield.Field

        A scalar field with angles. In the given direction the number of cells
        is reduced by one compared to the given field.

    Raises
    ------
    ValueError

        If ``field`` is not a vector field, or ``direction`` or ``units`` is
        invalid.

    Examples
    --------
    1. Computing the angle between neighbouring cells in z-direction.

    >>> import discretisedfield as df
    >>> import discretisedfield.tools as dft
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (100, 100, 100)
    >>> n = (10, 10, 10)
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
    >>> field = df.Field(mesh, dim=3, value=(0, 1, 0))
    ...
    >>> dft.neigbouring_cell_angle(field, direction='z')
    Field(...)

    """
    if not field.dim == 3:
        msg = f'Cannot compute spin angles for a field with {field.dim=}.'
        raise ValueError(msg)

    if direction not in dfu.axesdict.keys():
        msg = f'Cannot compute spin angles for direction {direction=}.'
        raise ValueError(msg)

    if units not in ['rad', 'deg']:
        msg = f'Units {units=} not supported.'
        raise ValueError(msg)

    # Orientation field
    fo = field.orientation

    if direction == 'x':
        dot_product = np.einsum('...j,...j->...',
                                fo.array[:-1, ...],
                                fo.array[1:, ...])
        delta_p = np.divide((field.mesh.dx, 0, 0), 2)

    elif direction == 'y':
        dot_product = np.einsum('...j,...j->...',
                                fo.array[:, :-1, ...],
                                fo.array[:, 1:, ...])
        delta_p = np.divide((0, field.mesh.dy, 0), 2)

    elif direction == 'z':
        dot_product = np.einsum('...j,...j->...',
                                fo.array[..., :-1, :],
                                fo.array[..., 1:, :])
        delta_p = np.divide((0, 0, field.mesh.dz), 2)

    # Define new mesh.
    p1 = np.add(field.mesh.region.pmin, delta_p)
    p2 = np.subtract(field.mesh.region.pmax, delta_p)
    mesh = df.Mesh(p1=p1, p2=p2, cell=field.mesh.cell)

    angles = np.arccos(np.clip(dot_product, -1.0, 1.0))
    if units == 'deg':
        angles = np.degrees(angles)

    return df.Field(mesh, dim=1, value=angles.reshape(*angles.shape, 1))


def max_neigbouring_cell_angle(field, /, units='rad'):
    """Calculate maximum angle between neighbouring cells in all directions.

    This function computes an angle between a cell and all its six neighbouring
    cells and assigns the maximum to that cell. The calculation is only
    possible for vector fields (``dim=3``). Angles are computed in degrees if
    ``units='deg'`` and in radians if ``units='rad'``.

    The resulting field has one discretisation cell less in the specified
    direction.

    Parameters
    ----------
    field : discretisedfield.Field

        Vector field.

    units : str, optional

        Angles are computed in degrees if ``units='deg'`` and in radians if
        ``units='rad'``. Defaults to ``'rad'``.

    Returns
    -------
    discretisedfield.Field

        A scalar field with maximum angles.

    Examples
    --------
    1. Computing the maximum angle between neighbouring cells.

    >>> import discretisedfield as df
    >>> import discretisedfield.tools as dft
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (100, 100, 100)
    >>> n = (10, 10, 10)
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
    >>> field = df.Field(mesh, dim=3, value=(0, 1, 0))
    ...
    >>> dft.max_neigbouring_cell_angle(field)
    Field(...)

    """
    x_angles = neigbouring_cell_angle(field, 'x', units=units).array.squeeze()
    y_angles = neigbouring_cell_angle(field, 'y', units=units).array.squeeze()
    z_angles = neigbouring_cell_angle(field, 'z', units=units).array.squeeze()

    max_angles = np.zeros((*field.array.shape[:-1], 6), dtype=np.float64)
    max_angles[1:, :, :, 0] = x_angles
    max_angles[:-1, :, :, 1] = x_angles
    max_angles[:, 1:, :, 2] = y_angles
    max_angles[:, :-1, :, 3] = y_angles
    max_angles[:, :, 1:, 4] = z_angles
    max_angles[:, :, :-1, 5] = z_angles
    max_angles = max_angles.max(axis=-1, keepdims=True)

    return df.Field(field.mesh, dim=1, value=max_angles)


def count_large_cell_angle_regions(field, /, min_angle,
                                   direction=None, units='rad'):
    """Count regions with large angles between neighbouring cells.

    This method counts regions, where the angle between neighbouring
    cells is above the given threshold. If ``direction`` is not specified
    the maximum of all neighbouring cells is used, otherwise only neighbouring
    cells in the given direction are taken into account. The minimum angle can
    be specified both in radians and degrees, depending on ``units``.

    Parameters
    ----------
    field : discretisedfield.Field

        Vector field.

    min_angle : numbers.Real

        Minimum angle to count. Can be either radians or degrees depending
        on ``units``.

    direction : str, optional

        Direction of neighbouring cells. Can be ``None`` or one of ``x``,
        ``y``, or ``z``. If ``None``, all directions are taken into account.
        Defaults to ``None``.

    units : str, optional

        Unit of ``min_angle``. Can be ``rad`` for radians or ``deg`` for
        degrees. Defaults to ``rad``.

    Returns
    -------
    int

        Number of regions.

    Examples
    --------
    1. Counting regions depending on all directions.

    >>> import discretisedfield as df
    >>> import discretisedfield.tools as dft
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (100, 100, 100)
    >>> n = (10, 10, 10)
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
    >>> field = df.Field(mesh, dim=3, \
                         value=lambda p: (0, 0, 1) if p[0] < 50 \
                         else (0, 0, -1))
    ...
    >>> dft.count_large_cell_angle_regions(field, min_angle=90, units='deg')
    1

    2. Counting regions depending on a single direction.

    >>> import discretisedfield as df
    >>> import discretisedfield.tools as dft
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (100, 100, 100)
    >>> n = (10, 10, 10)
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
    >>> field = df.Field(mesh, dim=3, \
                         value=lambda p: (0, 0, 1) if p[0] < 50 \
                                                   else (0, 0, -1))
    ...
    >>> dft.count_large_cell_angle_regions(field, min_angle=90, units='deg', \
                                           direction='x')
    1
    >>> dft.count_large_cell_angle_regions(field, min_angle=90, units='deg', \
                                           direction='y')
    0
    """
    if direction is None:
        cell_angles = max_neigbouring_cell_angle(field, units=units).array
    else:
        cell_angles = neigbouring_cell_angle(field, direction=direction,
                                             units=units).array
    _, num_features = ndimage.label(cell_angles > min_angle)
    return num_features


def count_bps(field, /, direction='x'):
    """Bloch point count and arrangement.

    Function to obtain information about Bloch point number and arrangement.
    The calculations are based on emergent magnetic field. The normalised
    volume integral over subvolumes, increasing cell by cell in the given
    ``direction`` is computed to obtain the local number of Bloch points at
    each point in the given ``direction``. Bloch point count and arangement
    are obtained by summing jumps in the local number of Bloch points.

    The results are:

    - Total number of Bloch points.
    - Number of head-to-head Bloch points.
    - Number of tail-to-tail Bloch points.
    - Arrangement of Bloch points in the given ``direction``. Starting from the
      lower end the local Bloch point count and the number of cells over which
      it stays constant are reported.

    Parameters
    ----------
    field : discretisedfield.Field

        Vector field.

    direction : str, optional

        Direction in which to compute arrangement. Can be one of ``x``, ``y``,
        or ``z``. Defaults to ``x``.

    Returns
    -------
    dict

        Dictionary containing information about BPs.

    Examples
    --------
    """
    F_div = emergent_magnetic_field(field.orientation).div

    d_vals = {'x': df.dx, 'y': df.dy, 'z': df.dz}
    averaged = str.replace('xyz', direction, '')
    dF = d_vals[averaged[0]] * d_vals[averaged[1]]
    dl = d_vals[direction]

    F_red = df.integral(F_div * dF, direction=averaged)
    F_int = df.integral(F_red * dl, direction=direction, improper=True)
    bp_number = (F_int / (4 * np.pi)).array.squeeze().round()
    bp_count = bp_number[1:] - bp_number[:-1]

    results = {}
    results['bp_number'] = abs(bp_count).sum()
    results['bp_number_hh'] = abs(bp_count[bp_count < 0].sum())
    results['bp_number_tt'] = bp_count[bp_count > 0].sum()

    # pattern = list([<local BP_count>, <repetitions>])
    pattern = [[bp_number[0], 1]]
    for q_val in bp_number[1:]:
        if q_val == pattern[-1][0]:
            pattern[-1][1] += 1
        else:
            pattern.append([q_val, 1])
    results[f'bp_pattern_{direction}'] = str(pattern)

    return results


def _demag_tensor_field_based(mesh):
    """Fourier transform of the demag tensor.

    Computes the demag tensor in Fourier space. Only the six different
    components Nxx, Nyy, Nzz, Nxy, Nxz, Nyz are returned.

    This version is using discretisedfield which makes it easy to understand
    but slow compared to the numpy version. (The reason is the array
    initialisation which is basically a large for-loop.) For actual use the
    numpy version should be used. This version is kept as a reference.

    Parameters
    ----------
    mesh : discretisedfield.Mesh
        Mesh to compute the demag tensor on.

    Returns
    -------
    discretisedfield.Field
        Demag tensor in Fourier space.

    """
    p1 = [(-i + 1) * j - j/2 for i, j in zip(mesh.n, mesh.cell)]
    p2 = [(i - 1) * j + j/2 for i, j in zip(mesh.n, mesh.cell)]
    n = [2*i - 1 for i in mesh.n]
    mesh_new = df.Mesh(p1=p1, p2=p2, n=n)

    return df.Field(mesh_new, dim=6, value=_N(mesh_new),
                    components=['xx', 'yy', 'zz', 'xy', 'xz', 'yz']).fftn


def demag_tensor(mesh):
    """Fourier transform of the demag tensor.

    Computes the demag tensor in Fourier space. Only the six different
    components Nxx, Nyy, Nzz, Nxy, Nxz, Nyz are returned.

    Parameters
    ----------
    mesh : discretisedfield.Mesh
        Mesh to compute the demag tensor on.

    Returns
    -------
    discretisedfield.Field
        Demag tensor in Fourier space.
    """
    x = np.linspace((-mesh.n[0] + 1) * mesh.cell[0],
                    (mesh.n[0] - 1) * mesh.cell[0], mesh.n[0] * 2 - 1)
    y = np.linspace((-mesh.n[1] + 1) * mesh.cell[1],
                    (mesh.n[1] - 1) * mesh.cell[1], mesh.n[1] * 2 - 1)
    z = np.linspace((-mesh.n[2] + 1) * mesh.cell[2],
                    (mesh.n[2] - 1) * mesh.cell[2], mesh.n[2] * 2 - 1)
    xx, yy, zz = np.meshgrid(x, y, z, indexing='ij')

    values = np.stack(_N(mesh)((xx, yy, zz)), axis=3)

    p1 = [(-i + 1) * j - j/2 for i, j in zip(mesh.n, mesh.cell)]
    p2 = [(i - 1) * j + j/2 for i, j in zip(mesh.n, mesh.cell)]
    n = [2*i - 1 for i in mesh.n]
    mesh_new = df.Mesh(p1=p1, p2=p2, n=n)

    return df.Field(mesh_new, dim=6, value=values,
                    components=['xx', 'yy', 'zz', 'xy', 'xz', 'yz']).fftn


def demag_field(m, tensor):
    """Demagnetisation field computed using ...

    Parameters
    ----------
    m : discretisedfield.Field
        Magnetisation field

    tensor : discretisedfield.field
        Demagnetisation tensor obatained with ``dft.demag_tensor``

    Returns
    -------
    discretisedfield.Field
        Demagnetisation field
    """
    m_pad = m.pad({d: (0, m.mesh.n[i] - 1)
                   for d, i in zip(['x', 'y', 'z'], range(3))},
                  mode='constant')
    m_fft = m_pad.fftn

    hx_fft = tensor.xx * m_fft.x + tensor.xy * m_fft.y + tensor.xz * m_fft.z
    hy_fft = tensor.xy * m_fft.x + tensor.yy * m_fft.y + tensor.yz * m_fft.z
    hz_fft = tensor.xz * m_fft.x + tensor.yz * m_fft.y + tensor.zz * m_fft.z

    H = (hx_fft << hy_fft << hz_fft).ifftn
    return df.Field(m.mesh, dim=3,
                    value=H.array[m.mesh.n[0] - 1:,
                                  m.mesh.n[1] - 1:,
                                  m.mesh.n[2] - 1:,
                                  :]).real


def _f(x, y, z):
    eps = 1e-18  # to avoid zero division; similar to Albert et al. 2015
    return (abs(y) / 2 * (z**2 - x**2)
            * np.arcsinh(abs(y) / (np.sqrt(x**2 + z**2) + eps))
            + abs(z) / 2 * (y**2 - x**2)
            * np.arcsinh(abs(z) / (np.sqrt(x**2 + y**2) + eps))
            - abs(x * y * z)
            * np.arctan(abs(y * z) /
                        (abs(x) * np.sqrt(x**2 + y**2 + z**2) + eps))
            + 1/6 * (2 * x**2 - y**2 - z**2) * np.sqrt(x**2 + y**2 + z**2))


def _g(x, y, z):
    eps = 1e-18  # to avoid zero division; similar to Albert et al. 2015
    return (x * y * z * np.arcsinh(z / (np.sqrt(x**2 + y**2) + eps))
            + y / 6 * (3 * z**2 - y**2)
            * np.arcsinh(x / (np.sqrt(y**2 + z**2) + eps))
            + x / 6 * (3 * z**2 - x**2)
            * np.arcsinh(y / (np.sqrt(x**2 + z**2) + eps))
            - z**3 / 6
            * np.arctan(x * y / (z * np.sqrt(x**2 + y**2 + z**2) + eps))
            - z * y**2 / 2
            * np.arctan(x * z / (y * np.sqrt(x**2 + y**2 + z**2) + eps))
            - z * x**2 / 2
            * np.arctan(y * z / (x * np.sqrt(x**2 + y**2 + z**2) + eps))
            - x * y * np.sqrt(x**2 + y**2 + z**2) / 3)


def _N_element(x, y, z, mesh, function):
    dx, dy, dz = mesh.cell
    value = 0.
    for i in np.rollaxis(np.indices((2,) * 6), 0, 7).reshape(64, 6):
        value += (-1)**np.sum(i) * function(x + (i[0] - i[3]) * dx,
                                            y + (i[1] - i[4]) * dy,
                                            z + (i[2] - i[5]) * dz)
    return - value / (4 * np.pi * np.prod(mesh.cell))


def _N(mesh):
    def _inner(p):
        x, y, z = p
        return (_N_element(x, y, z, mesh, _f),  # Nxx
                _N_element(y, z, x, mesh, _f),  # Nyy
                _N_element(z, x, y, mesh, _f),  # Nzz
                _N_element(x, y, z, mesh, _g),  # Nxy
                _N_element(x, z, y, mesh, _g),  # Nxz
                _N_element(y, z, x, mesh, _g))  # Nyz
    return _inner



import ubermagutil.units as uu
import matplotlib.pyplot as plt
import discretisedfield.util as dfu


class MplRegion:
    def __init__(self, region):
        self.region = region

    def __call__(self, *,
                 ax=None,
                 figsize=None,
                 multiplier=None,
                 color=dfu.cp_hex[0],
                 filename=None,
                 **kwargs):
        r"""``matplotlib`` plot.

        If ``ax`` is not passed, ``matplotlib.axes.Axes`` object is created
        automatically and the size of a figure can be specified using
        ``figsize``. The colour of lines depicting the region can be specified
        using ``color`` argument, which must be a valid ``matplotlib`` color.
        The plot is saved in PDF-format if ``filename`` is passed.

        It is often the case that the object size is either small (e.g. on a
        nanoscale) or very large (e.g. in units of kilometers). Accordingly,
        ``multiplier`` can be passed as :math:`10^{n}`, where :math:`n` is a
        multiple of 3 (..., -6, -3, 0, 3, 6,...). According to that value, the
        axes will be scaled and appropriate units shown. For instance, if
        ``multiplier=1e-9`` is passed, all axes will be divided by
        :math:`1\\,\\text{nm}` and :math:`\\text{nm}` units will be used as
        axis labels. If ``multiplier`` is not passed, the best one is
        calculated internally.

        This method is based on ``matplotlib.pyplot.plot``, so any keyword
        arguments accepted by it can be passed (for instance, ``linewidth``,
        ``linestyle``, etc.).

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional

            Axes to which the plot is added. Defaults to ``None`` - axes are
            created internally.

        figsize : (2,) tuple, optional

            The size of a created figure if ``ax`` is not passed. Defaults to
            ``None``.

        color : int, str, tuple, optional

            A valid ``matplotlib`` color for lines depicting the region.
            Defaults to the default color palette.

        multiplier : numbers.Real, optional

            Axes multiplier. Defaults to ``None``.

        filename : str, optional

            If filename is passed, the plot is saved. Defaults to ``None``.

        Examples
        --------
        1. Visualising the region using ``matplotlib``.

        >>> import discretisedfield as df
        ...
        >>> p1 = (-50e-9, -50e-9, 0)
        >>> p2 = (50e-9, 50e-9, 10e-9)
        >>> region = df.Region(p1=p1, p2=p2)
        >>> region.mpl()

        """
        ax = self._setup_axes(ax, figsize, projection='3d')

        multiplier = self._setup_multiplier(multiplier)

        kwargs.setdefault('color', color)

        rescaled_region = self.region / multiplier

        dfu.plot_box(ax=ax,
                     pmin=rescaled_region.pmin,
                     pmax=rescaled_region.pmax, **kwargs)

        self._axis_labels(ax, multiplier)

        # Overwrite default plotting parameters.
        ax.set_facecolor('#ffffff')  # white face color
        ax.tick_params(axis='both', which='major', pad=0)  # no pad for ticks

        if filename is not None:
            plt.savefig(filename, bbox_inches='tight', pad_inches=0)

    def _setup_axes(self, ax, figsize, **kwargs):
        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111, **kwargs)

        return ax

    def _setup_multiplier(self, multiplier):
        return self.region.multiplier if multiplier is None else multiplier

    def _axis_labels(self, ax, multiplier):
        unit = (rf' ({uu.rsi_prefixes[multiplier]}{self.region.unit})')
        ax.set(xlabel=f'x {unit}', ylabel=f'y {unit}', zlabel=f'z {unit}')



import numbers
import numpy as np
import discretisedfield as df


class DValue:
    """Infinitesimaly small value (differential).

    This class is used for defining infinitesimaly small values used for
    computing integrals. For instance ``dV``, ``dx``, surface vector field
    ``dS``, etc.

    Parameters
    ----------
    function : Python function

        A function which takes ``discretisedfield.Field`` object as an input
        and extracts the required differential.

    Examples
    --------
    1. Defining ``dV``.

    >>> import discretisedfield as df
    ...
    >>> dV = DValue(lambda f: f.mesh.dV)
    ...
    >>> p1 = (-50, -25, 0)
    >>> p2 = (50, 25, 5)
    >>> cell = (5, 5, 5)
    >>> region = df.Region(p1=p1, p2=p2)
    >>> mesh = df.Mesh(region=region, cell=cell)
    >>> field = df.Field(mesh, dim=1, value=3.14)
    ...
    >>> dV(field)
    125

    """
    def __init__(self, function, /):
        self.function = function

    def __call__(self, field):
        """Call ``self.function`` on ``field``.

        Parameters
        ----------
        field : discretisedfield.Field

            Field object.

        Returns
        -------
        numbers.Real, discretisedfield.Field

            Inifinitesimaly small value.

        Example
        -------
        1. Computing ``dx``.

        >>> import discretisedfield as df
        ...
        >>> dx = DValue(lambda f: f.mesh.dx)
        ...
        >>> p1 = (-50, -25, 0)
        >>> p2 = (50, 25, 5)
        >>> cell = (5, 5, 5)
        >>> region = df.Region(p1=p1, p2=p2)
        >>> mesh = df.Mesh(region=region, cell=cell)
        >>> field = df.Field(mesh, dim=1, value=3.14)
        ...
        >>> dx(field)
        5

        """
        return self.function(field)

    def __abs__(self):
        """Absolute value (norm) operator.

        This method computes norm if differential is a vector field (``dS``).

        Returns
        -------
        DValue

            Result.

        Examples
        --------
        1. Computing absolute value (norm).

        >>> import discretisedfield as df
        ...
        >>> dS = DValue(lambda f: f.mesh.dS)
        ...
        >>> p1 = (-50, -25, 0)
        >>> p2 = (50, 25, 5)
        >>> cell = (5, 5, 5)
        >>> region = df.Region(p1=p1, p2=p2)
        >>> mesh = df.Mesh(region=region, cell=cell)
        >>> field = df.Field(mesh, dim=3, value=(1, -9, 2))
        ...
        >>> abs(dS)(field.plane('z')).average
        25.0

        """
        return self.__class__(lambda f: abs(self(f)))

    def __mul__(self, other):
        """Binary ``*`` operator.

        It can be applied between:

        1. Two ``DValue`` objects.

        2. A field of any dimension and ``DValue`` object,

        3. ``DValue`` object and ``numbers.Real``.

        Parameters
        ----------
        other : DValue, discretisedfield.Field, numbers.Real

            Second operand.

        Returns
        -------
        DValue, discretisedfield.Field

            Result.

        Raises
        ------
        TypeError

            If the operator cannot be applied.

        Examples
        --------
        1. Multiplication examples.

        >>> import discretisedfield as df
        ...
        >>> dx = DValue(lambda f: f.mesh.dx)
        >>> dy = DValue(lambda f: f.mesh.dy)
        >>> dz = DValue(lambda f: f.mesh.dz)
        >>> dV = DValue(lambda f: f.mesh.dV)
        ...
        >>> p1 = (-50, -25, 0)
        >>> p2 = (50, 25, 5)
        >>> cell = (5, 5, 5)
        >>> region = df.Region(p1=p1, p2=p2)
        >>> mesh = df.Mesh(region=region, cell=cell)
        >>> field = df.Field(mesh, dim=3, value=(1, -9, 2))
        ...
        >>> # DValue multiplication
        >>> dxdydz = dx * dy * dz
        >>> # DValue and field multiplication
        >>> field * dV == field * dxdydz
        True
        >>> # Dvalue and int multiplication
        >>> field * dV * 2 == field * dxdydz * 2
        True

        """
        if isinstance(other, self.__class__):
            return self.__class__(lambda f: self(f) * other(f))
        elif isinstance(other, df.Field):
            return other * self
        elif isinstance(other, numbers.Real):
            return self.__class__(lambda f: self(f) * other)
        else:
            msg = (f'Unsupported operand type(s) for *: '
                   f'{type(self)=} and {type(other)=}.')
            raise TypeError(msg)

    def __rmul__(self, other):
        return self * other

    def __matmul__(self, other):
        """Binary ``@`` operator, defined as a dot product.

        It can be applied between:

        1. Two ``DValue`` objects.

        2. A vector field and ``DValue`` object,

        3. ``DValue`` object and ``array_like`` length-3 object.

        Parameters
        ----------
        other : DValue, discretisedfield.Field, (3,) array_like

            Second operand.

        Returns
        -------
        DValue, discretisedfield.Field

            Result.

        Raises
        ------
        TypeError

            If the operator cannot be applied.

        Example
        -------
        1. Dot product examples.

        >>> import discretisedfield as df
        ...
        >>> dS = DValue(lambda f: f.mesh.dS)
        ...
        >>> p1 = (-50, -25, 0)
        >>> p2 = (50, 25, 5)
        >>> cell = (5, 5, 5)
        >>> region = df.Region(p1=p1, p2=p2)
        >>> mesh = df.Mesh(region=region, cell=cell)
        >>> field = df.Field(mesh, dim=3, value=(1, -9, 2))
        ...
        >>> # DValue dot product
        >>> dS2 = dS * dS
        >>> # DValue and field dot product.
        >>> field.plane('z') @ dS
        Field(...)
        >>> # DValue and array_like object
        >>> field.plane('z') * (dS @ (1, 1, 1))
        Field(...)

        """
        if isinstance(other, self.__class__):
            return self.__class__(lambda f: self(f) @ other(f))
        elif isinstance(other, df.Field):
            return other @ self
        elif isinstance(other, (list, tuple, np.ndarray)):
            return self.__class__(lambda f: self(f) @ other)
        else:
            msg = (f'Unsupported operand type(s) for *: '
                   f'{type(self)=} and {type(other)=}.')
            raise TypeError(msg)

    def __rmatmul__(self, other):
        return self @ other


def integral(field, direction='xyz', improper=False):
    """Integral.

    This function calls ``integral`` method of the ``discrteisedfield.Field``
    object.

    For details, please refer to
    :py:func:`~discretisedfield.Field.integral`

    """
    return field.integral(direction=direction, improper=improper)


dx = DValue(lambda f: f.mesh.dx)
dy = DValue(lambda f: f.mesh.dy)
dz = DValue(lambda f: f.mesh.dz)
dV = DValue(lambda f: f.mesh.dV)
dS = DValue(lambda f: f.mesh.dS)



"""Matplotlib-based plotting."""
import warnings

import mpl_toolkits.axes_grid1.inset_locator
import matplotlib.pyplot as plt
import numpy as np

import discretisedfield.util as dfu
import ubermagutil.units as uu


class Mpl:
    """Matplotlib-based plotting convenience methods.

    Parameters
    ----------
    data : df.Field

        Field sliced with a plane, e.g. field.plane('x').

    """

    def __init__(self, field):
        # We never allow plotting fields that are not sliced.
        if not field.mesh.attributes['isplane']:
            msg = 'The field must be sliced before it can be plotted.'
            raise ValueError(msg)

        self.field = field  # TODO: Consider renaming data to field.
        self.planeaxis = dfu.raxesdict[field.mesh.attributes['planeaxis']]

    def __call__(self,
                 ax=None,
                 figsize=None,
                 multiplier=None,
                 scalar_kwargs=None,
                 vector_kwargs=None,
                 filename=None):
        """Plot the field on a plane.

        This is a convenience method used for quick plotting, and it combines
        ``discretisedfield.plotting.Mpl.scalar`` and
        ``discretisedfield.plotting.Mpl.vector`` methods. Depending on the
        dimensionality of the field's value, it automatically determines what
        plot is going to be shown. For a scalar field, only
        ``discretisedfield.plotting.Mpl.scalar`` is used, whereas for a vector
        field, both ``discretisedfield.plotting.Mpl.scalar`` and
        ``discretisedfield.plotting.Mpl.vector`` plots are shown so that vector
        plot visualises the in-plane components of the vector and scalar plot
        encodes the out-of-plane component.

        All the default values can be changed by passing arguments, which are
        then used in subplots. The way parameters of this function are used to
        create plots can be understood with the following code snippet.
        ``scalar_field`` and ``vector_field`` are computed internally (based on
        the dimension of the field).

        .. code-block::

            if ax is None:
                fig = plt.figure(figsize=figsize)
                ax = fig.add_subplot(111)

            if scalar_field is not None:
                scalar_field.mpl.scalar(ax=ax, multiplier=multiplier,
                                        **scalar_kwargs)
            if vector_field is not None:
                vector_field.mpl.vector(ax=ax, multiplier=multiplier,
                                        **vector_kwargs)

            if filename is not None:
                plt.savefig(filename, bbox_inches='tight', pad_inches=0.02)
            ```

            Therefore, to understand the meaning of the keyword arguments which
            can be passed to this method, please refer to
            ``discretisedfield.plotting.Mpl.scalar`` and
            ``discretisedfield.plotting.Mpl.vector`` documentation.

        Example
        -------
        .. plot:: :context: close-figs

            1. Visualising the field using ``matplotlib``.

            >>> import discretisedfield as df
            ...
            >>> p1 = (0, 0, 0)
            >>> p2 = (100, 100, 100)
            >>> n = (10, 10, 10)
            >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
            >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
            >>> field.plane(z=50, n=(5, 5)).mpl()

        .. seealso::

            :py:func:`~discretisedfield.plotting.Mpl.scalar`
            :py:func:`~discretisedfield.plotting.Mpl.vector`

        """
        ax = self._setup_axes(ax, figsize)

        multiplier = self._setup_multiplier(multiplier)

        scalar_kwargs = {} if scalar_kwargs is None else scalar_kwargs
        vector_kwargs = {} if vector_kwargs is None else vector_kwargs
        vector_kwargs.setdefault('use_color', False)
        vector_kwargs.setdefault('colorbar', False)

        # Set up default scalar and vector fields.
        # TODO: Do we allow user to specify what scalar and vector fields are?
        # Did we have that before?
        if self.field.dim == 1:
            scalar_field = self.field
            vector_field = None

        elif self.field.dim == 2:
            scalar_field = None
            vector_field = self.field

        else:
            vector_field = self.field
            scalar_field = getattr(self.field, self.planeaxis)
            scalar_kwargs['colorbar_label'] = scalar_kwargs.get(
                'colorbar_label', f'{self.planeaxis}-component')

        # TODO user should specify filter_field=None to avoid filtering
        # TODO what is the norm if dim=1
        scalar_kwargs.setdefault('filter_field', self.field.norm)

        if scalar_field is not None:
            scalar_field.mpl.scalar(ax=ax, multiplier=multiplier,
                                    **scalar_kwargs)
        if vector_field is not None:
            vector_field.mpl.vector(ax=ax, multiplier=multiplier,
                                    **vector_kwargs)

        self._axis_labels(ax, multiplier)

        if filename is not None:
            plt.savefig(filename, bbox_inches='tight', pad_inches=0.02)

    def scalar(self,
               ax=None,
               figsize=None,
               multiplier=None,
               filter_field=None,
               colorbar=True,
               colorbar_label=None,  # TODO: Can we maybe use an empty string?
               filename=None,
               symmetric_clim=False,
               **kwargs):
        r"""Plot the scalar field on a plane.

        Before the field can be plotted, it must be sliced with a plane (e.g.
        ``field.plane('z')``). In addition, field must be a scalar field
        (``dim=1``). Otherwise, ``ValueError`` is raised. ``mpl_scalar`` adds
        the plot to ``matplotlib.axes.Axes`` passed via ``ax`` argument. If
        ``ax`` is not passed, ``matplotlib.axes.Axes`` object is created
        automatically and the size of a figure can be specified using
        ``figsize``. By passing ``filter_field`` the points at which the pixels
        are not coloured can be determined. More precisely, only those
        discretisation cells where ``filter_field != 0`` are plotted. Colorbar
        is shown by default and it can be removed from the plot by passing
        ``colorbar=False``. The label for the colorbar can be defined by
        passing ``colorbar_label`` as a string.

        It is often the case that the region size is small (e.g. on a
        nanoscale) or very large (e.g. in units of kilometers). Accordingly,
        ``multiplier`` can be passed as :math:`10^{n}`, where :math:`n` is a
        multiple of 3  (..., -6, -3, 0, 3, 6,...). According to that value, the
        axes will be scaled and appropriate units shown. For instance, if
        ``multiplier=1e-9`` is passed, all mesh points will be divided by
        :math:`1\,\text{nm}` and :math:`\text{nm}` units will be used as
        axis labels. If ``multiplier`` is not passed, the best one is
        calculated internally. The plot can be saved as a PDF when ``filename``
        is passed.

        This method plots the field using ``matplotlib.pyplot.imshow``
        function, so any keyword arguments accepted by it can be passed (for
        instance, ``cmap`` - colormap, ``clim`` - colorbar limits, etc.).

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional

            Axes to which the field plot is added. Defaults to ``None`` - axes
            are created internally.

        figsize : (2,) tuple, optional

            The size of a created figure if ``ax`` is not passed. Defaults to
            ``None``.

        filter_field : discretisedfield.Field, optional

            A scalar field used for determining whether certain discretisation
            cells are coloured. More precisely, only those discretisation cells
            where ``filter_field != 0`` are plotted. Defaults to ``None``.

        colorbar : bool, optional

            If ``True``, colorbar is shown and it is hidden when ``False``.
            Defaults to ``True``.

        colorbar_label : str, optional

            Colorbar label. Defaults to ``None``.

        multiplier : numbers.Real, optional

            ``multiplier`` can be passed as :math:`10^{n}`, where :math:`n` is
            a multiple of 3 (..., -6, -3, 0, 3, 6,...). According to that
            value, the axes will be scaled and appropriate units shown. For
            instance, if ``multiplier=1e-9`` is passed, the mesh points will be
            divided by :math:`1\\,\\text{nm}` and :math:`\\text{nm}` units will
            be used as axis labels. Defaults to ``None``.

        filename : str, optional

            If filename is passed, the plot is saved. Defaults to ``None``.

        Raises
        ------
        ValueError

            If the field has not been sliced, its dimension is not 1, or the
            dimension of ``filter_field`` is not 1.

        Example
        -------
        .. plot::
            :context: close-figs

            1. Visualising the scalar field using ``matplotlib``.

            >>> import discretisedfield as df
            ...
            >>> p1 = (0, 0, 0)
            >>> p2 = (100, 100, 100)
            >>> n = (10, 10, 10)
            >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
            >>> field = df.Field(mesh, dim=1, value=2)
            ...
            >>> field.plane('y').mpl.scalar()

        .. seealso:: :py:func:`~discretisedfield.plotting.Mpl.vector`

        """
        if self.field.dim > 1:
            msg = f'Cannot plot {self.field.dim=} field.'
            raise ValueError(msg)

        ax = self._setup_axes(ax, figsize)

        multiplier = self._setup_multiplier(multiplier)

        points, values = map(list, zip(*list(self.field)))

        pmin = np.divide(self.field.mesh.region.pmin, multiplier)
        pmax = np.divide(self.field.mesh.region.pmax, multiplier)

        # TODO: After refactoring code, maybe extent and n can become
        # part of PlaneMesh.
        extent = [pmin[self.field.mesh.attributes['axis1']],
                  pmax[self.field.mesh.attributes['axis1']],
                  pmin[self.field.mesh.attributes['axis2']],
                  pmax[self.field.mesh.attributes['axis2']]]
        n = (self.field.mesh.n[self.field.mesh.attributes['axis2']],
             self.field.mesh.n[self.field.mesh.attributes['axis1']])

        values = self._filter_values(filter_field, points, values)

        if symmetric_clim and 'clim' not in kwargs.keys():
            vmin = np.min(values)
            vmax = np.max(values)
            if np.sign(vmin) != np.sign(vmax):
                vmax_abs = max(abs(vmin), vmax)
                kwargs['clim'] = (-vmax_abs, vmax_abs)
            else:
                warnings.warn('Symmetric clim only possible if the field has '
                              'positive and negative values.')
        cp = ax.imshow(np.array(values).reshape(n),
                       origin='lower', extent=extent, **kwargs)

        if colorbar:
            cbar = plt.colorbar(cp, ax=ax)
            if colorbar_label is not None:
                cbar.ax.set_ylabel(colorbar_label)

        self._axis_labels(ax, multiplier)

        if filename is not None:
            # TODO: We use pad inches 0 and 0.02.
            # We should figure out which one is the best.
            plt.savefig(filename, bbox_inches='tight', pad_inches=0)

    def lightness(self,
                  ax=None,
                  figsize=None,
                  multiplier=None,
                  filter_field=None,
                  lightness_field=None,
                  clim=None,
                  colorwheel=True,
                  colorwheel_xlabel=None,
                  colorwheel_ylabel=None,
                  colorwheel_args=None,
                  filename=None,
                  **kwargs):
        """Lightness plots.

        Uses HSV to show inplane angle and lightness for out-of-plane (3d) or
        norm (1d/2d) of the field. By passing a scalar field as
        ``lightness_field``, lightness component is can be specified
        independently of the field dimension. Most parameters are the same as
        for ``discretisedfield.plotting.Mpl.scalar``. Colormap cannot be passed
        using ``kwargs``. Instead of having a colorbar a ``colorwheel`` is
        displayed by default.

        The use of a ``filter_field`` is not yet supported (passing a
        ``filter_field`` has no effect).

        lightness_field : discretisedfield.Field, optional

            A scalar field used for adding lightness to the color. Field values
            are hue. Defaults to ``None``.

        colorwheel : bool, optional

            To control if a colorwheel is shown, defaults to ``True``.

        colorwheel_xlabel : str, optional

            If specified, the string and an arrow are plotted onto the
            colorwheel (in x-direction).

        colorwheel_ylabel : str, optional

            If specified, the string and an arrow are plotted onto the
            colorwheel (in y-direction).

        colorwheel_args : dict, optional

            Additional keyword arguments to pass to the colorwheel function.
            For details see ``discretisedfield.plotting.Mpl.colorwheel`` and
            ``mpl_toolkits.axes_grid1.inset_locator.inset_axes``.

        Example
        -------
        .. plot::
            :context: close-figs

            1. Visualising the scalar field using ``matplotlib``.

            >>> import discretisedfield as df
            ...
            >>> p1 = (0, 0, 0)
            >>> p2 = (100, 100, 100)
            >>> n = (10, 10, 10)
            >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
            >>> field = df.Field(mesh, dim=3, value=(1, 2, 3))
            ...
            >>> field.plane('z').mpl.lightness()

        """
        if self.field.dim == 2:
            if lightness_field is None:
                lightness_field = self.field.norm
            return self.field.angle.mpl.lightness(
                ax=ax,
                figsize=figsize,
                multiplier=multiplier,
                filter_field=filter_field,
                lightness_field=lightness_field,
                clim=clim,
                colorwheel=colorwheel,
                colorwheel_xlabel=colorwheel_xlabel,
                colorwheel_ylabel=colorwheel_ylabel,
                colorwheel_args=colorwheel_args,
                filename=filename,
                **kwargs)
        elif self.field.dim == 3:
            if lightness_field is None:
                lightness_field = getattr(self.field, self.planeaxis)
            return self.field.angle.mpl.lightness(
                ax=ax,
                figsize=figsize,
                multiplier=multiplier,
                filter_field=filter_field,
                lightness_field=lightness_field,
                clim=clim,
                colorwheel=colorwheel,
                colorwheel_xlabel=colorwheel_xlabel,
                colorwheel_ylabel=colorwheel_ylabel,
                colorwheel_args=colorwheel_args,
                filename=filename,
                **kwargs)

        ax = self._setup_axes(ax, figsize)

        multiplier = self._setup_multiplier(multiplier)

        points, values = map(list, zip(*list(self.field)))

        # TODO Requires refactoring of df.Mesh
        # rescaled_region = self.field.mesh.region.rescale(multiplier)
        # pmin = rescaled_region.pmin
        # pmax = rescaled_region.pmax

        pmin = np.divide(self.field.mesh.region.pmin, multiplier)
        pmax = np.divide(self.field.mesh.region.pmax, multiplier)

        # This depends on plane mesh orientation and should be in PlaneMesh.
        extent = [pmin[self.field.mesh.attributes['axis1']],
                  pmax[self.field.mesh.attributes['axis1']],
                  pmin[self.field.mesh.attributes['axis2']],
                  pmax[self.field.mesh.attributes['axis2']]]
        n = (self.field.mesh.n[self.field.mesh.attributes['axis2']],
             self.field.mesh.n[self.field.mesh.attributes['axis1']])

        if lightness_field is None:
            lightness_field = self.field.norm
        else:
            if lightness_field.dim != 1:
                msg = f'Cannot use {lightness_field.dim=} lightness_field.'
                raise ValueError(msg)

        lightness = [lightness_field(i) for i in self.field.mesh]

        # values = self._filter_values(filter_field, points, values)
        rgb = dfu.hls2rgb(hue=values,
                          lightness=lightness,
                          saturation=None,
                          lightness_clim=clim).squeeze()

        # TODO: filter field affects color range
        # rgb = self._filter_values(filter_field, points, rgb)
        # rgba = np.ones((len(rgb), 4))
        # rgba[..., :3] = rgb
        # rgba[..., 3][np.isnan(rgb[:, 0])] = 0

        kwargs['cmap'] = 'hsv'  # only hsv cmap allowed
        # TODO filtering
        # ax.imshow(rgba.reshape((*n, 4)), origin='lower',
        #           extent=extent, **kwargs)
        ax.imshow(rgb.reshape((*n, 3)), origin='lower',
                  extent=extent, **kwargs)
        if colorwheel:
            if colorwheel_args is None:
                colorwheel_args = {}
            cw_ax = add_colorwheel(ax, **colorwheel_args)
            if colorwheel_xlabel is not None:
                cw_ax.arrow(100, 100, 60, 0, width=5, fc='w', ec='w')
                cw_ax.annotate(colorwheel_xlabel, (115, 140), c='w')
            if colorwheel_ylabel is not None:
                cw_ax.arrow(100, 100, 0, -60, width=5, fc='w', ec='w')
                cw_ax.annotate(colorwheel_ylabel, (40, 80), c='w')

        self._axis_labels(ax, multiplier)

        if filename is not None:
            plt.savefig(filename, bbox_inches='tight', pad_inches=0)

    def vector(self,
               ax=None,
               figsize=None,
               multiplier=None,
               use_color=True,
               color_field=None,
               colorbar=True,
               colorbar_label=None,
               filename=None,
               **kwargs):
        r"""Plot the vector field on a plane.

        Before the field can be plotted, it must be sliced with a plane (e.g.
        ``field.plane('z')``). In addition, field must be a vector field
        (``dim=3``). Otherwise, ``valueerror`` is raised. ``mpl_vector`` adds
        the plot to ``matplotlib.axes.axes`` passed via ``ax`` argument. If
        ``ax`` is not passed, ``matplotlib.axes.axes`` object is created
        automatically and the size of a figure can be specified using
        ``figsize``. By default, plotted vectors are coloured according to the
        out-of-plane component of the vectors. This can be changed by passing
        ``color_field`` with ``dim=1``. To disable colouring of the plot,
        ``use_color=false`` can be passed. A uniform vector colour can be
        obtained by specifying ``color=color`` which is passed to matplotlib
        and ``use_color=false``. Colorbar is shown by default and it can
        be removed from the plot by passing ``colorbar=false``. The label for
        the colorbar can be defined by passing ``colorbar_label`` as a string.
        It is often the case that the region size is small (e.g. on a
        nanoscale) or very large (e.g. in units of kilometers). accordingly,
        ``multiplier`` can be passed as :math:`10^{n}`, where :math:`n` is a
        multiple of 3  (..., -6, -3, 0, 3, 6,...). according to that value, the
        axes will be scaled and appropriate units shown. For instance, if
        ``multiplier=1e-9`` is passed, all mesh points will be divided by
        :math:`1\\,\\text{nm}` and :math:`\\text{nm}` units will be used as
        axis labels. If ``multiplier`` is not passed, the best one is
        calculated internally. The plot can be saved as a pdf when ``filename``
        is passed.

        This method plots the field using ``matplotlib.pyplot.quiver``
        function, so any keyword arguments accepted by it can be passed (for
        instance, ``cmap`` - colormap, ``clim`` - colorbar limits, etc.). In
        particular, there are cases when ``matplotlib`` fails to find optimal
        scale for plotting vectors. More precisely, sometimes vectors appear
        too large in the plot. This can be resolved by passing ``scale``
        argument, which scales all vectors in the plot. In other words, larger
        ``scale``, smaller the vectors and vice versa. Please note that scale
        can be in a very large range (e.g. 1e20).

        parameters
        ----------
        ax : matplotlib.axes.axes, optional

            Axes to which the field plot is added. defaults to ``None`` - axes
            are created internally.

        figsize : tuple, optional

            The size of a created figure if ``ax`` is not passed. defaults to
            ``None``.

        color_field : discretisedfield.field, optional

            A scalar field used for colouring the vectors. defaults to ``None``
            and vectors are coloured according to their out-of-plane
            components.

        colorbar : bool, optional

            If ``true``, colorbar is shown and it is hidden when ``false``.
            defaults to ``true``.

        colorbar_label : str, optional

            Colorbar label. Defaults to ``None``.

        multiplier : numbers.real, optional

            ``multiplier`` can be passed as :math:`10^{n}`, where :math:`n` is
            a multiple of 3 (..., -6, -3, 0, 3, 6,...). According to that
            value, the axes will be scaled and appropriate units shown. For
            instance, if ``multiplier=1e-9`` is passed, the mesh points will be
            divided by :math:`1\\,\\text{nm}` and :math:`\\text{nm}` units will
            be used as axis labels. defaults to ``None``.

        filename : str, optional

            If filename is passed, the plot is saved. defaults to ``None``.

        raises
        ------
        valueerror

            If the field has not been sliced, its dimension is not 3, or the
            dimension of ``color_field`` is not 1.

        example
        -------
        .. plot::
            :context: close-figs

            1. visualising the vector field using ``matplotlib``.

            >>> import discretisedfield as df
            ...
            >>> p1 = (0, 0, 0)
            >>> p2 = (100, 100, 100)
            >>> n = (10, 10, 10)
            >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
            >>> field = df.Field(mesh, dim=3, value=(1.1, 2.1, 3.1))
            ...
            >>> field.plane('y').mpl.vector()

        .. seealso:: :py:func:`~discretisedfield.field.mpl_scalar`

        """
        if self.field.dim not in [2, 3]:
            msg = f'cannot plot dim={self.field.dim} field.'
            raise ValueError(msg)

        ax = self._setup_axes(ax, figsize)

        multiplier = self._setup_multiplier(multiplier)

        points, values = map(list, zip(*list(self.field)))

        # remove points and values where norm is 0.
        points = [p for p, v in zip(points, values)
                  if not np.equal(v, 0).all()]
        values = [v for v in values if not np.equal(v, 0).all()]

        if use_color:
            if color_field is None:
                # todo raises an exception by default; options:
                # - warning + automatically specify use_color=false
                # - use_color=false as default
                if self.field.dim == 2:
                    msg = 'automatic coloring is only supported for 3d fields.'
                    raise ValueError(msg)
                color_field = getattr(self.field, self.planeaxis)

            colors = [color_field(p) for p in points]

        # "unpack" values inside arrays and convert to np.ndarray.
        points = np.array(list(zip(*points)))
        values = np.array(list(zip(*values)))

        points = np.divide(points, multiplier)

        if use_color:
            cp = ax.quiver(points[self.field.mesh.attributes['axis1']],
                           points[self.field.mesh.attributes['axis2']],
                           values[self.field.mesh.attributes['axis1']],
                           values[self.field.mesh.attributes['axis2']],
                           colors, pivot='mid', **kwargs)
        else:
            ax.quiver(points[self.field.mesh.attributes['axis1']],
                      points[self.field.mesh.attributes['axis2']],
                      values[self.field.mesh.attributes['axis1']],
                      values[self.field.mesh.attributes['axis2']],
                      pivot='mid', **kwargs)

        ax.set_aspect('equal')
        if colorbar and use_color:
            cbar = plt.colorbar(cp, ax=ax)
            if colorbar_label is not None:
                cbar.ax.set_ylabel(colorbar_label)

        self._axis_labels(ax, multiplier)

        if filename is not None:
            plt.savefig(filename, bbox_inches='tight', pad_inches=0)

    def contour(self,
                ax=None,
                figsize=None,
                multiplier=None,
                filter_field=None,
                colorbar=True,
                colorbar_label=None,
                filename=None,
                **kwargs):
        """Contour line plot.

        See ``discretisedfield.plotting.Mpl.scalar`` for details and meaning of
        the arguments.

        Example
        -------

        .. plot::
            :context: close-figs

            1. Visualising the scalar field using ``matplotlib``.

            >>> import discretisedfield as df
            >>> import math
            ...
            >>> p1 = (0, 0, 0)
            >>> p2 = (100, 100, 100)
            >>> n = (10, 10, 10)
            >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
            >>> def init_value(point):
            ...     x, y, z = point
            ...     return math.sin(x) + math.sin(y)
            >>> field = df.Field(mesh, dim=1, value=init_value)
            ...
            >>> field.plane('z').mpl.contour()

        """
        if self.field.dim != 1:
            msg = f'Cannot plot dim={self.field.dim} field.'
            raise ValueError(msg)

        ax = self._setup_axes(ax, figsize)

        multiplier = self._setup_multiplier(multiplier)

        points, values = map(list, zip(*list(self.field)))

        values = self._filter_values(filter_field, points, values)

        n = (self.field.mesh.n[self.field.mesh.attributes['axis2']],
             self.field.mesh.n[self.field.mesh.attributes['axis1']])

        points = np.array(list(zip(*points)))
        points = np.divide(points, multiplier)

        values = np.array(values).reshape(n)

        cp = ax.contour(points[self.field.mesh.attributes['axis1']].reshape(n),
                        points[self.field.mesh.attributes['axis2']].reshape(n),
                        values, **kwargs)
        ax.set_aspect('equal')

        if colorbar:
            cbar = plt.colorbar(cp, ax=ax)
            if colorbar_label is not None:
                cbar.ax.set_ylabel(colorbar_label)

        self._axis_labels(ax, multiplier)

        if filename is not None:
            plt.savefig(filename, bbox_inches='tight', pad_inches=0)

    def _setup_axes(self, ax, figsize, **kwargs):
        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111, **kwargs)

        return ax

    def _setup_multiplier(self, multiplier):
        return (self.field.mesh.region.multiplier
                if multiplier is None else multiplier)

    def _filter_values(self, filter_field, points, values):
        if filter_field is None:
            return values

        if filter_field.dim != 1:
            msg = f'Cannot use {filter_field.dim=}.'
            raise ValueError(msg)

        return [values[i] if filter_field(point) != 0 else np.nan
                for i, point in enumerate(points)]

    def _axis_labels(self, ax, multiplier):
        unit = (rf' ({uu.rsi_prefixes[multiplier]}'
                rf'{self.field.mesh.attributes["unit"]})')
        ax.set_xlabel(dfu.raxesdict[self.field.mesh.attributes['axis1']]
                      + unit)
        ax.set_ylabel(dfu.raxesdict[self.field.mesh.attributes['axis2']]
                      + unit)

    def __dir__(self):
        dirlist = dir(self.__class__)

        for attr in ['vector'] if self.field.dim == 1 else ['scalar',
                                                            'contour']:
            dirlist.remove(attr)

        return dirlist


def add_colorwheel(ax, width=1, height=1, loc='lower right', **kwargs):
    """Colorwheel for hsv plots.

    Creates colorwheel on new inset axis. See
    ``mpl_toolkits.axes_grid1.inset_locator.inset_axes`` for the meaning of the
    arguments and other possible keyword arguments.

    Example
    -------

    .. plot::
        :context: close-figs

        1. Adding a colorwheel to an empty axis
        >>> import discretisedfield.plotting as dfp
        >>> import matplotlib.pyplot as plt
        ...
        >>> fig, ax = plt.subplots()
        >>> ins_ax = dfp.add_colorwheel(ax)

    """
    n = 200
    x = np.linspace(-1, 1, n)
    y = np.linspace(-1, 1, n)
    X, Y = np.meshgrid(x, y)

    theta = np.arctan2(Y, X)
    r = np.sqrt(X**2 + Y**2)

    rgb = dfu.hls2rgb(hue=theta, lightness=r,
                      lightness_clim=[0, 1 / np.sqrt(2)])

    theta = theta.reshape((n, n, 1))

    rgba = np.zeros((n, n, 4))
    for i, xi in enumerate(x):
        for j, yi in enumerate(y):
            if xi**2 + yi**2 <= 1:
                rgba[i, j, :3] = rgb[i, j, :]
                rgba[i, j, 3] = 1

    ax_ins = mpl_toolkits.axes_grid1.inset_locator.inset_axes(
        ax, width=width, height=height, loc=loc, **kwargs)
    ax_ins.imshow(rgba[:, ::-1, :])
    ax_ins.axis('off')
    return ax_ins



import os
import pytest
import discretisedfield as df
import discretisedfield.tools as dft


def test_topological_charge():
    p1 = (0, 0, 0)
    p2 = (10, 10, 10)
    cell = (2, 2, 2)
    mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

    # f(x, y, z) = (0, 0, 0)
    # -> Q(f) = 0
    f = df.Field(mesh, dim=3, value=(0, 0, 0))

    for method in ['continuous', 'berg-luescher']:
        q = dft.topological_charge_density(f.plane('z'), method=method)

        assert q.dim == 1
        assert q.average == 0
        for absolute in [True, False]:
            assert dft.topological_charge(f.plane('z'),
                                          method=method,
                                          absolute=absolute) == 0

    # Skyrmion (with PBC) from a file
    test_filename = os.path.join(os.path.dirname(__file__),
                                 'test_sample/',
                                 'skyrmion.omf')
    f = df.Field.fromfile(test_filename)

    for method in ['continuous', 'berg-luescher']:
        q = dft.topological_charge_density(f.plane('z'), method=method)

        assert q.dim == 1
        assert q.average > 0
        for absolute in [True, False]:
            Q = dft.topological_charge(f.plane('z'),
                                       method=method,
                                       absolute=absolute)
            assert abs(Q) < 1 and abs(Q - 1) < 0.15

    # Not sliced
    f = df.Field(mesh, dim=3, value=(1, 2, 3))
    for function in ['topological_charge_density', 'topological_charge']:
        for method in ['continuous', 'berg-luescher']:
            with pytest.raises(ValueError):
                res = getattr(dft, function)(f, method=method)

    # Scalar field
    f = df.Field(mesh, dim=1, value=3.14)
    for function in ['topological_charge_density', 'topological_charge']:
        for method in ['continuous', 'berg-luescher']:
            with pytest.raises(ValueError):
                res = getattr(dft, function)(f.plane('z'), method=method)

    # Method does not exist
    f = df.Field(mesh, dim=3, value=(1, 2, 3))
    for function in ['topological_charge_density', 'topological_charge']:
        with pytest.raises(ValueError):
            res = getattr(dft, function)(f.plane('z'), method='wrong')


def test_emergent_magnetic_field():
    p1 = (0, 0, 0)
    p2 = (10, 10, 10)
    cell = (2, 2, 2)
    mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

    # f(x, y, z) = (0, 0, 0)
    # -> F(f) = 0
    f = df.Field(mesh, dim=3, value=(0, 0, 0))

    assert dft.emergent_magnetic_field(f).dim == 3
    assert dft.emergent_magnetic_field(f).average == (0, 0, 0)

    with pytest.raises(ValueError):
        res = dft.emergent_magnetic_field(f.x)


def test_neigbouring_cell_angle():
    p1 = (0, 0, 0)
    p2 = (100, 100, 100)
    n = (10, 10, 10)
    mesh = df.Mesh(p1=p1, p2=p2, n=n)
    field = df.Field(mesh, dim=3, value=(0, 1, 0))

    for direction in 'xyz':
        for units in ['rad', 'deg']:
            sa = dft.neigbouring_cell_angle(field,
                                            direction=direction,
                                            units=units)
            assert sa.average == 0

    # Exceptions
    scalar_field = df.Field(mesh, dim=1, value=5)
    with pytest.raises(ValueError):
        res = dft.neigbouring_cell_angle(scalar_field, direction='x')

    with pytest.raises(ValueError):
        res = dft.neigbouring_cell_angle(field, direction='l')

    with pytest.raises(ValueError):
        res = dft.neigbouring_cell_angle(field, direction='x', units='wrong')


def test_max_neigbouring_cell_angle():
    p1 = (0, 0, 0)
    p2 = (100, 100, 100)
    n = (10, 10, 10)
    mesh = df.Mesh(p1=p1, p2=p2, n=n)
    field = df.Field(mesh, dim=3, value=(0, 1, 0))

    for units in ['rad', 'deg']:
        assert dft.max_neigbouring_cell_angle(field, units=units).average == 0


def test_count_lange_cell_angle_regions():
    p1 = (0, 0, 0)
    p2 = (10, 10, 10)
    n = (10, 10, 10)
    ps1 = (3, 3, 0)
    ps2 = (6, 4, 10)
    subregions = {'sub': df.Region(p1=ps1, p2=ps2)}
    mesh = df.Mesh(p1=p1, p2=p2, n=n, subregions=subregions)
    field = df.Field(mesh, dim=3, value={'sub': (0, 0, 1),
                                         'default': (0, 0, -1)})

    for direction, res in [['x', 2], ['y', 1], ['z', 0]]:
        assert dft.count_large_cell_angle_regions(
            field, min_angle=1, direction=direction) == res

    assert dft.count_large_cell_angle_regions(field, min_angle=1) == 1


def test_count_bps():
    # TODO use BP field from file
    p1 = (0, 0, 0)
    p2 = (10, 10, 10)
    n = (10, 10, 10)
    mesh = df.Mesh(p1=p1, p2=p2, n=n)
    field = df.Field(mesh, dim=3, value=(0, 0, 1))

    result = dft.count_bps(field)
    assert result['bp_number'] == 0
    assert result['bp_number_hh'] == 0
    assert result['bp_number_tt'] == 0
    assert result['bp_pattern_x'] == '[[0.0, 10]]'



"""Matplotlib based plotting."""
from discretisedfield.plotting.mpl import Mpl, add_colorwheel
from discretisedfield.plotting.mpl_region import MplRegion
from discretisedfield.plotting.k3d import K3d
from discretisedfield.plotting.k3d_region import K3dRegion
